@@ -20,6 +20,7 @@
 #include <ctype.h>
 #include "propsel.h"
 #include "util.h"
+#include "unaligned.h"
 
 #include "mpath_persist.h"
 #include "mpathpr.h"
@@ -30,7 +31,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
-#include <sys/resource.h>
 
 #define __STDC_FORMAT_MACROS 1
 
@@ -47,15 +47,7 @@ mpath_lib_init (void)
 		return NULL;
 	}
 
-	if (conf->max_fds) {
-		struct rlimit fd_limit;
-
-		fd_limit.rlim_cur = conf->max_fds;
-		fd_limit.rlim_max = conf->max_fds;
-		if (setrlimit(RLIMIT_NOFILE, &fd_limit) < 0)
-			condlog(0, "can't set open fds limit to %d : %s",
-				   conf->max_fds, strerror(errno));
-	}
+	set_max_fds(conf->max_fds);
 
 	return conf;
 }
@@ -335,6 +327,7 @@ int mpath_persistent_reserve_out ( int fd, int rq_servact, int rq_scope,
 
 	conf = get_multipath_config();
 	select_reservation_key(conf, mpp);
+	select_all_tg_pt(conf, mpp);
 	put_multipath_config(conf);
 
 	memcpy(&prkey, paramp->sa_key, 8);
@@ -343,7 +336,8 @@ int mpath_persistent_reserve_out ( int fd, int rq_servact, int rq_scope,
 	      rq_servact == MPATH_PROUT_REG_SA) ||
 	     rq_servact == MPATH_PROUT_REG_IGN_SA)) {
 		memcpy(&mpp->reservation_key, paramp->sa_key, 8);
-		if (update_prkey(alias, get_be64(mpp->reservation_key))) {
+		if (update_prkey_flags(alias, get_be64(mpp->reservation_key),
+				       paramp->sa_flags)) {
 			condlog(0, "%s: failed to set prkey for multipathd.",
 				alias);
 			ret = MPATH_PR_DMMP_ERROR;
@@ -456,7 +450,7 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 	unsigned int rq_type, struct prout_param_descriptor * paramp, int noisy)
 {
 
-	int i, j;
+	int i, j, k;
 	struct pathgroup *pgp = NULL;
 	struct path *pp = NULL;
 	int rollback = 0;
@@ -464,11 +458,14 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 	int rc;
 	int count=0;
 	int status = MPATH_PR_SUCCESS;
+	int all_tg_pt;
 	uint64_t sa_key = 0;
 
 	if (!mpp)
 		return MPATH_PR_DMMP_ERROR;
 
+	all_tg_pt = (mpp->all_tg_pt == ALL_TG_PT_ON ||
+		     paramp->sa_flags & MPATH_F_ALL_TG_PT_MASK);
 	active_pathcount = pathcount(mpp, PATH_UP) + pathcount(mpp, PATH_GHOST);
 
 	if (active_pathcount == 0) {
@@ -476,16 +473,14 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 		return MPATH_PR_DMMP_ERROR;
 	}
 
-	if ( paramp->sa_flags & MPATH_F_ALL_TG_PT_MASK ) {
-		condlog (1, "Warning: ALL_TG_PT is set. Configuration not supported");
-	}
-
 	struct threadinfo thread[active_pathcount];
+	int hosts[active_pathcount];
 
 	memset(thread, 0, sizeof(thread));
 
 	/* init thread parameter */
 	for (i =0; i< active_pathcount; i++){
+		hosts[i] = -1;
 		thread[i].param.rq_servact = rq_servact;
 		thread[i].param.rq_scope = rq_scope;
 		thread[i].param.rq_type = rq_type;
@@ -493,7 +488,7 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 		thread[i].param.noisy = noisy;
 		thread[i].param.status = MPATH_PR_SKIP;
 
-		condlog (3, "THRED ID [%d] INFO]", i);
+		condlog (3, "THREAD ID [%d] INFO]", i);
 		condlog (3, "rq_servact=%d ", thread[i].param.rq_servact);
 		condlog (3, "rq_scope=%d ", thread[i].param.rq_scope);
 		condlog (3, "rq_type=%d ", thread[i].param.rq_type);
@@ -514,6 +509,16 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 				condlog (1, "%s: %s path not up. Skip.", mpp->wwid, pp->dev);
 				continue;
 			}
+			if (all_tg_pt && pp->sg_id.host_no != -1) {
+				for (k = 0; k < count; k++) {
+					if (pp->sg_id.host_no == hosts[k]) {
+						condlog(3, "%s: %s host %d matches skip.", pp->wwid, pp->dev, pp->sg_id.host_no);
+						break;
+					}
+				}
+				if (k < count)
+					continue;
+			}
 			strncpy(thread[count].param.dev, pp->dev,
 				FILE_NAME_SIZE - 1);
 
@@ -531,10 +536,12 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 				condlog (0, "%s: failed to create thread %d", mpp->wwid, rc);
 				thread[count].param.status = MPATH_PR_THREAD_ERROR;
 			}
+			else
+				hosts[count] = pp->sg_id.host_no;
 			count = count + 1;
 		}
 	}
-	for( i=0; i < active_pathcount ; i++){
+	for( i=0; i < count ; i++){
 		if (thread[i].param.status != MPATH_PR_THREAD_ERROR) {
 			rc = pthread_join(thread[i].id, NULL);
 			if (rc){
@@ -543,12 +550,7 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 		}
 		if (!rollback && (thread[i].param.status == MPATH_PR_RESERV_CONFLICT)){
 			rollback = 1;
-			sa_key = 0;
-			for (i = 0; i < 8; ++i){
-				if (i > 0)
-					sa_key <<= 8;
-				sa_key |= paramp->sa_key[i];
-			}
+			sa_key = get_unaligned_be64(&paramp->sa_key[0]);
 			status = MPATH_PR_RESERV_CONFLICT ;
 		}
 		if (!rollback && (status == MPATH_PR_SUCCESS)){
@@ -557,11 +559,10 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 	}
 	if (rollback && ((rq_servact == MPATH_PROUT_REG_SA) && sa_key != 0 )){
 		condlog (3, "%s: ERROR: initiating pr out rollback", mpp->wwid);
-		for( i=0 ; i < active_pathcount ; i++){
+		memcpy(&paramp->key, &paramp->sa_key, 8);
+		memset(&paramp->sa_key, 0, 8);
+		for( i=0 ; i < count ; i++){
 			if(thread[i].param.status == MPATH_PR_SUCCESS) {
-				memcpy(&thread[i].param.paramp->key, &thread[i].param.paramp->sa_key, 8);
-				memset(&thread[i].param.paramp->sa_key, 0, 8);
-				thread[i].param.status = MPATH_PR_SUCCESS;
 				rc = pthread_create(&thread[i].id, &attr, mpath_prout_pthread_fn,
 						(void *)(&thread[i].param));
 				if (rc){
@@ -571,7 +572,7 @@ int mpath_prout_reg(struct multipath *mpp,int rq_servact, int rq_scope,
 			} else
 				thread[i].param.status = MPATH_PR_SKIP;
 		}
-		for(i=0; i < active_pathcount ; i++){
+		for(i=0; i < count ; i++){
 			if (thread[i].param.status != MPATH_PR_SKIP &&
 			    thread[i].param.status != MPATH_PR_THREAD_ERROR) {
 				rc = pthread_join(thread[i].id, NULL);
@@ -720,7 +721,7 @@ int mpath_prout_rel(struct multipath *mpp,int rq_servact, int rq_scope,
 		}
 	}
 	pthread_attr_destroy (&attr);
-	for (i = 0; i < active_pathcount; i++){
+	for (i = 0; i < count; i++){
 		if (thread[i].param.status != MPATH_PR_THREAD_ERROR) {
 			rc = pthread_join (thread[i].id, NULL);
 			if (rc){
@@ -729,7 +730,7 @@ int mpath_prout_rel(struct multipath *mpp,int rq_servact, int rq_scope,
 		}
 	}
 
-	for (i = 0; i < active_pathcount; i++){
+	for (i = 0; i < count; i++){
 		/*  check thread status here and return the status */
 
 		if (thread[i].param.status == MPATH_PR_RESERV_CONFLICT)
@@ -909,7 +910,7 @@ int update_map_pr(struct multipath *mpp)
 
 	if (resp->prin_descriptor.prin_readkeys.additional_length == 0 )
 	{
-		condlog(0,"%s: No key found. Device may not be registered. ", mpp->alias);
+		condlog(3,"%s: No key found. Device may not be registered. ", mpp->alias);
 		free(resp);
 		return MPATH_PR_SUCCESS;
 	}
