@@ -26,14 +26,13 @@
 #define SCSI_COMMAND_TERMINATED	0x22
 #define SG_ERR_DRIVER_SENSE	0x08
 #define RECOVERED_ERROR		0x01
+#define ILLEGAL_REQUEST		0x05
 
 
 #define CURRENT_PAGE_CODE_VALUES	0
 #define CHANGEABLE_PAGE_CODE_VALUES	1
 
-#define MSG_RDAC_UP    "rdac checker reports path is up"
-#define MSG_RDAC_DOWN  "rdac checker reports path is down"
-#define MSG_RDAC_GHOST "rdac checker reports path is ghost"
+#define MSG_RDAC_DOWN  " reports path is down"
 #define MSG_RDAC_DOWN_TYPE(STR) MSG_RDAC_DOWN": "STR
 
 #define RTPG_UNAVAILABLE	0x3
@@ -139,11 +138,6 @@ void libcheck_free (struct checker * c)
 	return;
 }
 
-void libcheck_repair (struct checker * c)
-{
-	return;
-}
-
 static int
 do_inq(int sg_fd, unsigned int pg_op, void *resp, int mx_resp_len,
        unsigned int timeout)
@@ -169,14 +163,14 @@ retry:
 	io_hdr.sbp = sense_b;
 	io_hdr.timeout = timeout * 1000;
 
-	if (ioctl(sg_fd, SG_IO, &io_hdr) < 0)
-		return 1;
+	if (ioctl(sg_fd, SG_IO, &io_hdr) < 0 && errno == ENOTTY)
+		return PATH_WILD;
 
 	/* treat SG_ERR here to get rid of sg_err.[ch] */
 	io_hdr.status &= 0x7e;
 	if ((0 == io_hdr.status) && (0 == io_hdr.host_status) &&
 	    (0 == io_hdr.driver_status))
-		return 0;
+		return PATH_UP;
 
 	/* check if we need to retry this error */
 	if (io_hdr.info & SG_INFO_OK_MASK) {
@@ -205,10 +199,14 @@ retry:
 			else
 				sense_key = sense_buffer[2] & 0xf;
 			if (RECOVERED_ERROR == sense_key)
-				return 0;
+				return PATH_UP;
+			else if (ILLEGAL_REQUEST == sense_key)
+				return PATH_WILD;
+			condlog(1, "rdac checker: INQUIRY failed with sense key %02x",
+				sense_key);
 		}
 	}
-	return 1;
+	return PATH_DOWN;
 }
 
 struct volume_access_inq
@@ -224,41 +222,69 @@ struct volume_access_inq
 	char dontcare1[34];
 };
 
-const char
-*checker_msg_string(struct volume_access_inq *inq)
+enum {
+	RDAC_MSGID_NOT_CONN = CHECKER_FIRST_MSGID,
+	RDAC_MSGID_IN_STARTUP,
+	RDAC_MSGID_NON_RESPONSIVE,
+	RDAC_MSGID_IN_RESET,
+	RDAC_MSGID_FW_DOWNLOADING,
+	RDAC_MSGID_QUIESCED,
+	RDAC_MSGID_SERVICE_MODE,
+	RDAC_MSGID_UNAVAILABLE,
+	RDAC_MSGID_INQUIRY_FAILED,
+};
+
+#define _IDX(x) (RDAC_MSGID_##x - CHECKER_FIRST_MSGID)
+const char *libcheck_msgtable[] = {
+	[_IDX(NOT_CONN)] = MSG_RDAC_DOWN_TYPE("lun not connected"),
+	[_IDX(IN_STARTUP)] = MSG_RDAC_DOWN_TYPE("ctlr is in startup sequence"),
+	[_IDX(NON_RESPONSIVE)] =
+	MSG_RDAC_DOWN_TYPE("non-responsive to queries"),
+	[_IDX(IN_RESET)] = MSG_RDAC_DOWN_TYPE("ctlr held in reset"),
+	[_IDX(FW_DOWNLOADING)] =
+	MSG_RDAC_DOWN_TYPE("ctlr firmware downloading"),
+	[_IDX(QUIESCED)] = MSG_RDAC_DOWN_TYPE("ctlr quiesced by admin request"),
+	[_IDX(SERVICE_MODE)] = MSG_RDAC_DOWN_TYPE("ctlr is in service mode"),
+	[_IDX(UNAVAILABLE)] = MSG_RDAC_DOWN_TYPE("ctlr is unavailable"),
+	[_IDX(INQUIRY_FAILED)] = MSG_RDAC_DOWN_TYPE("inquiry failed"),
+	NULL,
+};
+
+static int
+checker_msg_string(const struct volume_access_inq *inq)
 {
 	/* lun not connected */
 	if (((inq->PQ_PDT & 0xE0) == 0x20) || (inq->PQ_PDT & 0x7f))
-		return MSG_RDAC_DOWN_TYPE("lun not connected");
+		return RDAC_MSGID_NOT_CONN;
 
 	/* if no tpg data is available, give the generic path down message */
 	if (!(inq->avtcvp & 0x10))
-		return MSG_RDAC_DOWN;
+		return CHECKER_MSGID_DOWN;
 
 	/* controller is booting up */
 	if (((inq->aas_cur & 0x0F) == RTPG_TRANSITIONING) &&
 		(inq->aas_alt & 0x0F) != RTPG_TRANSITIONING)
-		return MSG_RDAC_DOWN_TYPE("ctlr is in startup sequence");
+		return RDAC_MSGID_IN_STARTUP;
 
 	/* if not unavailable, give generic message */
 	if ((inq->aas_cur & 0x0F) != RTPG_UNAVAILABLE)
-		return MSG_RDAC_DOWN;
+		return CHECKER_MSGID_DOWN;
 
 	/* target port group unavailable */
 	switch (inq->vendor_specific_cur) {
 	case RTPG_UNAVAIL_NON_RESPONSIVE:
-		return MSG_RDAC_DOWN_TYPE("non-responsive to queries");
+		return RDAC_MSGID_NON_RESPONSIVE;
 	case RTPG_UNAVAIL_IN_RESET:
-		return MSG_RDAC_DOWN_TYPE("ctlr held in reset");
+		return RDAC_MSGID_IN_RESET;
 	case RTPG_UNAVAIL_CFW_DL1:
 	case RTPG_UNAVAIL_CFW_DL2:
-		return MSG_RDAC_DOWN_TYPE("ctlr firmware downloading");
+		return RDAC_MSGID_FW_DOWNLOADING;
 	case RTPG_UNAVAIL_QUIESCED:
-		return MSG_RDAC_DOWN_TYPE("ctlr quiesced by admin request");
+		return RDAC_MSGID_QUIESCED;
 	case RTPG_UNAVAIL_SERVICE_MODE:
-		return MSG_RDAC_DOWN_TYPE("ctlr is in service mode");
+		return RDAC_MSGID_SERVICE_MODE;
 	default:
-		return MSG_RDAC_DOWN_TYPE("ctlr is unavailable");
+		return RDAC_MSGID_UNAVAILABLE;
 	}
 }
 
@@ -269,12 +295,14 @@ int libcheck_check(struct checker * c)
 
 	inqfail = 0;
 	memset(&inq, 0, sizeof(struct volume_access_inq));
-	if (0 != do_inq(c->fd, 0xC9, &inq, sizeof(struct volume_access_inq),
-			c->timeout)) {
-		ret = PATH_DOWN;
+	ret = do_inq(c->fd, 0xC9, &inq, sizeof(struct volume_access_inq),
+		     c->timeout);
+	if (ret != PATH_UP) {
 		inqfail = 1;
 		goto done;
-	} else if (((inq.PQ_PDT & 0xE0) == 0x20) || (inq.PQ_PDT & 0x7f)) {
+	}
+
+	if (((inq.PQ_PDT & 0xE0) == 0x20) || (inq.PQ_PDT & 0x7f)) {
 		/* LUN not connected*/
 		ret = PATH_DOWN;
 		goto done;
@@ -311,15 +339,18 @@ int libcheck_check(struct checker * c)
 
 done:
 	switch (ret) {
+	case PATH_WILD:
+		c->msgid = CHECKER_MSGID_UNSUPPORTED;
+		break;
 	case PATH_DOWN:
-		MSG(c, "%s", (inqfail) ? MSG_RDAC_DOWN_TYPE("inquiry failed") :
-			checker_msg_string(&inq));
+		c->msgid = (inqfail ? RDAC_MSGID_INQUIRY_FAILED :
+			    checker_msg_string(&inq));
 		break;
 	case PATH_UP:
-		MSG(c, MSG_RDAC_UP);
+		c->msgid = CHECKER_MSGID_UP;
 		break;
 	case PATH_GHOST:
-		MSG(c, MSG_RDAC_GHOST);
+		c->msgid = CHECKER_MSGID_GHOST;
 		break;
 	}
 
