@@ -52,6 +52,7 @@
 #include "util.h"
 #include "config.h"
 #include "blacklist.h"
+#include "devmapper.h"
 
 #define MAX_ACCUMULATION_COUNT 2048
 #define MAX_ACCUMULATION_TIME 30*1000
@@ -104,29 +105,71 @@ uevq_cleanup(struct list_head *tmpq)
 	}
 }
 
-void
-uevent_get_wwid(struct uevent *uev)
+static const char* uevent_get_env_var(const struct uevent *uev,
+				      const char *attr)
 {
-	int i;
-	char *uid_attribute;
-	struct config * conf;
+	int i, len;
+	const char *p = NULL;
 
-	conf = get_multipath_config();
-	uid_attribute = parse_uid_attribute_by_attrs(conf->uid_attrs, uev->kernel);
-	put_multipath_config(conf);
+	if (attr == NULL)
+		goto invalid;
 
-	if (!uid_attribute)
-		return;
+	len = strlen(attr);
+	if (len == 0)
+		goto invalid;
 
 	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], uid_attribute, strlen(uid_attribute)) &&
-		    strlen(uev->envp[i]) > strlen(uid_attribute) &&
-		    uev->envp[i][strlen(uid_attribute)] == '=') {
-			uev->wwid = uev->envp[i] + strlen(uid_attribute) + 1;
+		const char *var = uev->envp[i];
+
+		if (strlen(var) > len &&
+		    !memcmp(var, attr, len) && var[len] == '=') {
+			p = var + len + 1;
 			break;
 		}
 	}
-	free(uid_attribute);
+
+	condlog(4, "%s: %s -> '%s'", __func__, attr, p);
+	return p;
+
+invalid:
+	condlog(2, "%s: empty variable name", __func__);
+	return NULL;
+}
+
+static int uevent_get_env_positive_int(const struct uevent *uev,
+				       const char *attr)
+{
+	const char *p = uevent_get_env_var(uev, attr);
+	char *q;
+	int ret;
+
+	if (p == NULL || *p == '\0')
+		return -1;
+
+	ret = strtoul(p, &q, 10);
+	if (*q != '\0' || ret < 0) {
+		condlog(2, "%s: invalid %s: '%s'", __func__, attr, p);
+		return -1;
+	}
+	return ret;
+}
+
+void
+uevent_get_wwid(struct uevent *uev)
+{
+	char *uid_attribute;
+	const char *val;
+	struct config * conf;
+
+	conf = get_multipath_config();
+	pthread_cleanup_push(put_multipath_config, conf);
+	uid_attribute = parse_uid_attribute_by_attrs(conf->uid_attrs, uev->kernel);
+	pthread_cleanup_pop(1);
+
+	val = uevent_get_env_var(uev, uid_attribute);
+	if (val)
+		uev->wwid = val;
+	FREE(uid_attribute);
 }
 
 bool
@@ -146,6 +189,7 @@ uevent_need_merge(void)
 bool
 uevent_can_discard(struct uevent *uev)
 {
+	int invalid = 0;
 	struct config * conf;
 
 	/*
@@ -157,13 +201,14 @@ uevent_can_discard(struct uevent *uev)
 	 * filter paths devices by devnode
 	 */
 	conf = get_multipath_config();
+	pthread_cleanup_push(put_multipath_config, conf);
 	if (filter_devnode(conf->blist_devnode, conf->elist_devnode,
-			   uev->kernel) > 0) {
-		put_multipath_config(conf);
-		return true;
-	}
-	put_multipath_config(conf);
+			   uev->kernel) > 0)
+		invalid = 1;
+	pthread_cleanup_pop(1);
 
+	if (invalid)
+		return true;
 	return false;
 }
 
@@ -217,7 +262,7 @@ merge_need_stop(struct uevent *earlier, struct uevent *later)
 	if (!earlier->wwid || !later->wwid)
 		return true;
 	/*
-	 * uevents merging stoped
+	 * uevents merging stopped
 	 * when we meet an opposite action uevent from the same LUN to AVOID
 	 * "add path1 |remove path1 |add path2 |remove path2 |add path3"
 	 * to merge as "remove path1, path2" and "add path1, path2, path3"
@@ -288,7 +333,7 @@ uevent_filter(struct uevent *later, struct list_head *tmpq)
 		 * by the later uevent
 		 */
 		if (uevent_can_filter(earlier, later)) {
-			condlog(2, "uevent: %s-%s has filtered by uevent: %s-%s",
+			condlog(3, "uevent: %s-%s has filtered by uevent: %s-%s",
 				earlier->kernel, earlier->action,
 				later->kernel, later->action);
 
@@ -312,7 +357,7 @@ uevent_merge(struct uevent *later, struct list_head *tmpq)
 		 * merge earlier uevents to the later uevent
 		 */
 		if (uevent_can_merge(earlier, later)) {
-			condlog(2, "merged uevent: %s-%s-%s with uevent: %s-%s-%s",
+			condlog(3, "merged uevent: %s-%s-%s with uevent: %s-%s-%s",
 				earlier->action, earlier->kernel, earlier->wwid,
 				later->action, later->kernel, later->wwid);
 
@@ -684,6 +729,12 @@ struct uevent *uevent_from_udev_device(struct udev_device *dev)
 		if (i == HOTPLUG_NUM_ENVP - 1)
 			break;
 	}
+	if (!uev->devpath || ! uev->action) {
+		udev_device_unref(dev);
+		condlog(1, "uevent missing necessary fields");
+		FREE(uev);
+		return NULL;
+	}
 	uev->udev = dev;
 	uev->envp[i] = NULL;
 
@@ -850,107 +901,52 @@ out:
 	return err;
 }
 
-int uevent_get_major(struct uevent *uev)
+int uevent_get_major(const struct uevent *uev)
 {
-	char *p, *q;
-	int i, major = -1;
-
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "MAJOR", 5) && strlen(uev->envp[i]) > 6) {
-			p = uev->envp[i] + 6;
-			major = strtoul(p, &q, 10);
-			if (p == q) {
-				condlog(2, "invalid major '%s'", p);
-				major = -1;
-			}
-			break;
-		}
-	}
-	return major;
+	return uevent_get_env_positive_int(uev, "MAJOR");
 }
 
-int uevent_get_minor(struct uevent *uev)
+int uevent_get_minor(const struct uevent *uev)
 {
-	char *p, *q;
-	int i, minor = -1;
-
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "MINOR", 5) && strlen(uev->envp[i]) > 6) {
-			p = uev->envp[i] + 6;
-			minor = strtoul(p, &q, 10);
-			if (p == q) {
-				condlog(2, "invalid minor '%s'", p);
-				minor = -1;
-			}
-			break;
-		}
-	}
-	return minor;
+	return uevent_get_env_positive_int(uev, "MINOR");
 }
 
-int uevent_get_disk_ro(struct uevent *uev)
+int uevent_get_disk_ro(const struct uevent *uev)
 {
-	char *p, *q;
-	int i, ro = -1;
-
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "DISK_RO", 7) && strlen(uev->envp[i]) > 8) {
-			p = uev->envp[i] + 8;
-			ro = strtoul(p, &q, 10);
-			if (p == q) {
-				condlog(2, "invalid read_only setting '%s'", p);
-				ro = -1;
-			}
-			break;
-		}
-	}
-	return ro;
+	return uevent_get_env_positive_int(uev, "DISK_RO");
 }
 
-char *uevent_get_dm_name(struct uevent *uev)
+static char *uevent_get_dm_str(const struct uevent *uev, char *attr)
 {
-	char *p = NULL;
-	int i;
+	const char *tmp = uevent_get_env_var(uev, attr);
 
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "DM_NAME", 7) &&
-		    strlen(uev->envp[i]) > 8) {
-			p = MALLOC(strlen(uev->envp[i] + 8) + 1);
-			strcpy(p, uev->envp[i] + 8);
-			break;
-		}
-	}
-	return p;
+	if (tmp == NULL)
+		return NULL;
+	return strdup(tmp);
 }
 
-char *uevent_get_dm_path(struct uevent *uev)
+char *uevent_get_dm_name(const struct uevent *uev)
 {
-	char *p = NULL;
-	int i;
-
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "DM_PATH", 7) &&
-		    strlen(uev->envp[i]) > 8) {
-			p = MALLOC(strlen(uev->envp[i] + 8) + 1);
-			strcpy(p, uev->envp[i] + 8);
-			break;
-		}
-	}
-	return p;
+	return uevent_get_dm_str(uev, "DM_NAME");
 }
 
-char *uevent_get_dm_action(struct uevent *uev)
+char *uevent_get_dm_path(const struct uevent *uev)
 {
-	char *p = NULL;
-	int i;
+	return uevent_get_dm_str(uev, "DM_PATH");
+}
 
-	for (i = 0; uev->envp[i] != NULL; i++) {
-		if (!strncmp(uev->envp[i], "DM_ACTION", 9) &&
-		    strlen(uev->envp[i]) > 10) {
-			p = MALLOC(strlen(uev->envp[i] + 10) + 1);
-			strcpy(p, uev->envp[i] + 10);
-			break;
-		}
-	}
-	return p;
+char *uevent_get_dm_action(const struct uevent *uev)
+{
+	return uevent_get_dm_str(uev, "DM_ACTION");
+}
+
+bool uevent_is_mpath(const struct uevent *uev)
+{
+	const char *uuid = uevent_get_env_var(uev, "DM_UUID");
+
+	if (uuid == NULL)
+		return false;
+	if (strncmp(uuid, UUID_PREFIX, UUID_PREFIX_LEN))
+		return false;
+	return uuid[UUID_PREFIX_LEN] != '\0';
 }
