@@ -1,3 +1,6 @@
+#ifndef _CLI_H_
+#define _CLI_H_
+
 #include <stdint.h>
 
 enum {
@@ -41,6 +44,7 @@ enum {
 	__SETPRKEY,
 	__UNSETPRKEY,
 	__KEY,
+	__LOCAL,
 };
 
 #define LIST		(1 << __LIST)
@@ -84,6 +88,7 @@ enum {
 #define SETPRKEY	(1ULL << __SETPRKEY)
 #define UNSETPRKEY	(1ULL << __UNSETPRKEY)
 #define KEY		(1ULL << __KEY)
+#define LOCAL		(1ULL << __LOCAL)
 
 #define INITIAL_REPLY_LEN	1200
 
@@ -126,3 +131,5 @@ void free_handlers (void);
 int cli_init (void);
 void cli_exit(void);
 char * key_generator (const char * str, int state);
+
+#endif /* _CLI_H_ */
